#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

double f,dt,t,pi2,amp,damp,scale;
long preblocks,blocks,postblocks,iok,isquare,hsquare,i,j,ii;
long dithering,payload,udchoice;

unsigned char pathname[512];
unsigned char leafname[64];
unsigned char outfilename[512];
unsigned char wtype,stype;
unsigned char command[512];
unsigned char filetag[64];

unsigned char appdirname[512];
unsigned char settingsfilename[512];
unsigned char userdatafilename[512];

unsigned char datablock[1152000];

long blocksize,icut,icut2,icut24,icut224,ibfactor,ok;


unsigned char ChunkID[5],Format[5],Subchunk1ID[5],Subchunk2ID[5];

long ChunkSize,Subchunk1Size,AudioFormat;
long NumChannels,SampleRate,ByteRate,BlockAlign;
long BitsPerSample,Subchunk2Size,BytesPerSample;

long ih,ibaseflag,samples_per_cycle,icycle,isc;
long i44,i48;


double s[129],c[129];

FILE* outfile;


void read_settings_file(void);

long even_test(long);
long ok_test(long);


void get_user_input(void);

void make_IQ_wave(void);
void load_IQ_wave(void);

void make_output_name(void);

void sinefix(void);
void squarefix(void);
void zero_block(void);
void save_block(void);

void make_sinewave(void);
void make_squarewave(void);
void make_trianglewave(void);
void make_impulsewave(void);
void make_userwave(void);
void load_values(double,double,long);
long quantise(double);
double dither(void);

void write_output_header(long);

void write4int(long);
void write2int(long);
void write4(unsigned char*);


int main(int arc,char*argv[])
{
  icut=32768; icut2=2*icut;  icut24=icut*256; icut224=2*icut24; 
  pi2=3.14159265359*2.0;
  
  printf("  WAVIQGenerator Ver 1.00 17 May 2011\n");
  printf("  ===================================\n\n");
  printf("This lets you generate LPCM WAV files containing IQ test waveforms.\n");
  printf("Please read the Help inside the application before first use\n");
  printf("and check you have the correct details in your Settings file!\n\n");
  
  sprintf(appdirname,"%s\0",argv[0]);
  ok=(long)strrchr(appdirname,'/')-(long)appdirname;
  appdirname[ok]=(char)0;
  printf("Application directory = %s\n",appdirname);
  sprintf(settingsfilename,"%s/%s\0",appdirname,"Settings\0");
  printf("Settings file at %s\n",settingsfilename);

  
  read_settings_file();
  get_user_input();
  
  if (ibaseflag==0)
  {
    printf("44.1k base rate\n");
    samples_per_cycle=i44;
  }
  else
  {
    printf("48k base rate\n");
    samples_per_cycle=i48;
  }
  
  printf(" Set to produce track with test waveform of %ld sec duration\n",blocks);
  printf("   with IQ waveform using %ld samples per base cycle\n\n",samples_per_cycle);
  
  payload=blocks*ByteRate;
  printf("Payload in Bytes = %ld\n",payload);
  
  make_IQ_wave();
  
  make_output_name();
  
  
  
  
  
  
  outfile=fopen(outfilename,"wd");
  
  write_output_header(payload);
  
  isc=0;
  
  j=0;
  while(j<blocks)
  {
    load_IQ_wave();
    save_block();
    j++;
    printf("%3ld\n",j);
  }
  
  
  fclose(outfile);
  printf("Track finished\n");
  
  
  
  
  printf("Done OK!\n");
}

void load_IQ_wave(void)
{
  /* This loads the cycle of cos and sine values into the current
     datablock. It allows for the cycle value to not have to fit an
     exact integer number of cycles into each block. It just
     carries on from each block to the next in the sequence  */
  
  long ims;
  double answerI,answerQ;
  ims=0; 
  do
  {
    answerI=s[isc];
    answerQ=c[isc];
    if(dithering==1)
    {
      answerI+=dither();
      answerQ+=dither();
    }
    load_values(answerI,answerQ,ims);
    ims++;
    isc++;
    if(isc==icycle)
    {
      isc=0;
    }
  } while (ims<SampleRate);
  
  printf("IQ wave loaded %ld\n",isc);
}

void make_IQ_wave(void)
{
  /* This generates the cycle of sine and cos values which can then
     be loaded into each block of data. It saves having to keep
     recalculating using sin and cos for each block. */
     
  double pcycle;
  
  icycle=samples_per_cycle*ih;
  
  f=((double)SampleRate)/icycle;
  printf("Frequency = %8.2lf Hz\n",f);
  printf(" using %ld samples per cycle\n",icycle);
  
  isc=0;
  do
  {
    pcycle=dt*(double)isc;
    pcycle=pi2*f*pcycle;
    s[isc]=sin(pcycle);
    c[isc]=cos(pcycle);
    
    printf("%4ld,  %7.4lf,  %7.4lf\n",isc,s[isc],c[isc]);
    
    
    isc++;
  } while (isc<icycle);
  
  
  isc=0;
  do
  {
    s[isc]=s[isc]*amp*scale;
    c[isc]=c[isc]*amp*scale;
    isc++;
  } while(isc<icycle);
  
  
  printf("IQ wave created\n");
}


void get_user_input(void)
{
  sprintf(ChunkID,"RIFF\0");
  sprintf(Format,"WAVE\0");
  sprintf(Subchunk1ID,"fmt \0");
  sprintf(Subchunk2ID,"data\0");
  
  Subchunk1Size=16;
  AudioFormat=1;
  NumChannels=2;
  
  printf("Choose bits per sample - Press\n");
  printf("  '1' for 16 bits per sample or\n");
  printf("  '2' for 24 bits per sample\n");
  printf("then press <enter> => ");
  scanf("%ld",&BitsPerSample);
  
  if (BitsPerSample==1)
  {
    BitsPerSample=16;
    BytesPerSample=2;
    scale=32767.0;
    ibfactor=4;
    BlockAlign=4;
  }
  else
  {
    BitsPerSample=24;
    BytesPerSample=3;
    scale=256.0*32767.0;
    ibfactor=6;
    BlockAlign=6;
  }
  
  printf("Choose sample rate - Press\n");
  printf("  '1' for  44.1k   '2' for  48k\n");
  printf("  '3' for  88.2k   '4' for  96k\n");
  printf("  '5' for 176.4k   '6' for 192k\n");
  printf("then press <enter> => ");
  scanf("%ld",&SampleRate);
  
  if(SampleRate<1)
  {
     SampleRate=1; printf("Out of range! Default to 44.1k\n");
  }
  if(SampleRate>6)
  {
     SampleRate=1; printf("Out of range! Default to 44.1k\n");
  }
  
  switch(SampleRate)
  {
    case 1:
      SampleRate=44100;   sprintf(filetag,"44k.wav\0");  ih=1; ibaseflag=0;
      break;
    case 2:
      SampleRate=48000;   sprintf(filetag,"48k.wav\0");  ih=1; ibaseflag=1;
      break;
    case 3:
      SampleRate=2*44100; sprintf(filetag,"88k.wav\0");  ih=2; ibaseflag=0;
      break;
    case 4:
      SampleRate=2*48000; sprintf(filetag,"96k.wav\0");  ih=2; ibaseflag=1;
      break;
    case 5:
      SampleRate=4*44100; sprintf(filetag,"176k.wav\0"); ih=4; ibaseflag=0;
      break;
    case 6:
      SampleRate=4*48000; sprintf(filetag,"192k.wav\0"); ih=4; ibaseflag=1;
  }
  
  
  blocksize=SampleRate*BytesPerSample*2;
  dt=1.0/(double)SampleRate;
  ByteRate=blocksize;
  
   
  printf("Choose to dither or not. \n Enter 1 for dither and 0 for undithered =>  ");
  scanf("%ld",&dithering);
  if (dithering!=1) dithering=0;
    
  if (dithering==1)
  {
    printf("Dithering OK\n");
  }
  else
  {
    printf("Undithered output\n");
  }
  
  printf("Enter amplitude [dbFS] => ");
  
  scanf("%lf",&damp);
  
  if (damp>0.0) damp=-damp;
  
  amp=damp/20.0;
  amp=pow(10.0,amp);
  printf("amplitude = %6.4lf wrt FS\n",amp);
  
  printf("Chosen values are:\n");
  printf(" Sample rate = %ld ksamples/sec per channel\n",SampleRate);
  printf("   Byte Rate = %ld bytes/sec  tag = %s\n",ByteRate,filetag);
  printf(" Sample interval %8.3lf microseconds\n",dt*1.0E6);
}




double dither(void)
{
  /* This procedure returns a 'random dither'
  levet as a real value in the range from -1.0
  to -1.0 with a triangular probability density */
  
  double answer;
  long dit;
  dit=rand()%128 + rand()%128;
  answer=(double)(dit-128);
  return answer/128.0;
}



long quantise(double rin)
{
  /* This proceedure takes a real value and
     returns a result as an integer which
     has been quantised into uniform steps. */
     
  long answer;
  
  if (rin>0.0)
  {
    answer=(long)(rin+0.5);
  }
  else
  {
    answer=(long)(rin-0.5);
  }
  return answer;
}




void save_block(void)
{
  fwrite(datablock,sizeof(unsigned char),blocksize,outfile);
}


void zero_block(void)
{
  long iz;
  iz=0;
  do
  {
    datablock[iz]=(unsigned char)0;
    iz++;
  } while (iz<blocksize);
  
}

void load_values(double vinI,double vinQ,long iv)
{
  
  /* This takes each stereo IQ pair of values, converts them
     into the correct sequence of bytes, and loads these into
     the data block.  */
  
  int ib,lt,rt;
  char l1,l2,l3,r1,r2,r3;
  
  ib=ibfactor*iv;
  
  if(vinI>scale)    vinI=scale;
  if(vinI<(-scale)) vinI=(-scale);
  
  if(vinQ>scale)    vinQ=scale;
  if(vinQ<(-scale)) vinQ=(-scale);
  
  lt=quantise(vinI);
  rt=quantise(vinQ);
  
  if(BytesPerSample==2)
  {
    if(lt<0) lt+=2*icut;
    if(rt<0) rt+=2*icut;
    
    l1=(char)((lt>>8)&255);
    l2=(char)(lt&255);
    r1=(char)((rt>>8)&255);
    r2=(char)(rt&255);
    
    datablock[ib]=l2;
    datablock[ib+1]=l1;
    datablock[ib+2]=r2;
    datablock[ib+3]=r1;
  }
  else
  {
    
    if(lt<0) lt+=2*icut24;
    if(rt<0) rt+=2*icut24;
    
    l1=(char)((lt>>16)&255);
    l2=(char)((lt>>8)&255);
    l3=(char)(lt&255);
    
    r1=(char)((rt>>16)&255);
    r2=(char)((rt>>8)&255);
    r3=(char)(rt&255);
    
    datablock[ib]=l3;
    datablock[ib+1]=l2;
    datablock[ib+2]=l1;
    
    datablock[ib+3]=r3;
    datablock[ib+4]=r2;
    datablock[ib+5]=r1;
    
  }
}


void sinefix(void)
{
  double fhfix;
  
  fhfix=(double)SampleRate;
  fhfix=fhfix*0.5;
  fhfix=fhfix-10.0;
  if (f<1.0) f=1.0;
  if (f>fhfix) f=fhfix;
  f=(double)(long)f;
  printf("Frequency (rounded to Hz) => %7.1lf Hz\n\n",f);
}

void squarefix(void)
{
  double ft,fdiff,fbest;
  long iseven,isok,ipass,ival;
  
  ival=1;
  fdiff=(double)SampleRate;
  
  
  do
  {
    iseven=even_test(ival);
    isok=ok_test(ival);
    ipass=iseven*isok;
    if(ipass==1)
    {
      ft=(double)SampleRate/(double)ival;
      if (fabs(ft-f)<fdiff)
      {
        fdiff=fabs(ft-f);
        fbest=ft;
        isquare=ival;
        hsquare=isquare/2;
      }
      
    }
    ival++;
  } while (ival<=SampleRate);
  
  
  
  f=fbest;
  
  
  printf("Correction for uniform rate to %ld samples per cycle\n",isquare);
  printf("Frequency %8.2lf Hz\n\n",f);
}


long even_test(long iet)
{
  long answer;
  if (iet%2==0)
  {
    answer=1;
  }
  else
  {
    answer=0;
  }
  return answer;
}


long ok_test(long iokt)
{
  long answer;
  if (SampleRate%iokt==0)
  {
    answer=1;
  }
  else
  {
    answer=0;
  }
  return answer;
}


void make_output_name(void)
{
 char ts[64];
  char du;
  long kpart,upart;
  
  if(dithering==1)
  {
    du='d';
  }
  else
  {
    du='u';
  }

    if (f>1000.0)
    {
      kpart=(long)(f/1000.0);
      upart=(long)f - kpart*1000;
      sprintf(ts,"%02ldk%03ldHz%03lddB\0",kpart,upart,(long)damp);
      printf("\na\n");
    }
    else
    {
      sprintf(ts,"%03ldHz%03lddB\0",(long)f,(long)damp);
      printf("\nb\n");
    }
    
    if(BytesPerSample==2)
    {
      sprintf(leafname,"IQ%s%c\0",ts,du);
    }
    else
    {
      sprintf(leafname,"IQ%s%c24bit\0",ts,du);
    }
    sprintf(outfilename,"%s/%s%s\0",pathname,leafname,filetag);
    printf("\nc\n");
  
  printf("Outfile file\n  leafname = %s\n  fullname = %s\n\n",leafname,outfilename);
  
}


void read_settings_file(void)
{
  /* Reads the settings file inside the application to allow
     some flexibility of the program settings */
     
  
  FILE* setfile;
  
  
  setfile=fopen(settingsfilename,"rt");
  
  fscanf(setfile,"%s",pathname);
  
  printf("Output file will be put into directory\n  = %s\n\n",pathname);
  
  fscanf(setfile,"%ld",&blocks);
  fscanf(setfile,"%ld",&i44);
  fscanf(setfile,"%ld",&i48);
  
  fclose(setfile);
  
  
  if(i44>32)
  {
    i44=32;
    printf("Setting for 44k cycle rate too large!\n  Set to 32\n");
  }
  
  
  if(i44<4)
  {
    i44=4;
    printf("Setting for 44k cycle rate too small!\n  Set to 4\n");
  }
  
  if(i48>32)
  {
    i48=32;
    printf("Setting for 44k cycle rate too large!\n  Set to 32\n");
  }
  
  if(i48<4)
  {
    i48=4;
    printf("Setting for 44k cycle rate too small!\n  Set to 4\n");
  }
  
  printf("Cycle base values %ld [44k1] and %ld [48k]\n",i44,i48);
  
  
  
  
}

void write_output_header(long plo)
{
  
  printf("Payload to write = %ld\n",plo);
  write4(ChunkID);
  write4int(plo+36);
  write4(Format);
  write4(Subchunk1ID);
  write4int(Subchunk1Size);
  write2int(AudioFormat);
  write2int(NumChannels);
  write4int(SampleRate);
  write4int(ByteRate);
  write2int(BlockAlign);
  write2int(BitsPerSample);
  write4(Subchunk2ID);
  write4int(plo);
  
}

void write4int(long inv)
{
  unsigned char cop1,cop2,cop3,cop4;
  
  cop1=(unsigned char)(inv&255);
  cop2=(unsigned char)((inv>>8)&255);
  cop3=(unsigned char)((inv>>16)&255);
  cop4=(unsigned char)((inv>>24)&255);
  
  fprintf(outfile,"%c",cop1);
  fprintf(outfile,"%c",cop2);
  fprintf(outfile,"%c",cop3);
  fprintf(outfile,"%c",cop4);
}


void write2int(long inv)
{
  unsigned char cop1,cop2;
  
  cop1=(unsigned char)(inv&255);
  cop2=(unsigned char)((inv>>8)&255);
  
  
  fprintf(outfile,"%c",cop1);
  fprintf(outfile,"%c",cop2);
  
}

void write4(unsigned char* outstring)
{
  long isp;
  
  isp=0;
  do
  {
    fprintf(outfile,"%c",outstring[isp]);
    isp++;
  } while (isp<4);
  
}


